/*
    Authors:
        Pavel Březina <pbrezina@redhat.com>

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#include "authselect_private.h"
#include "authselect_files.h"
#include "authselect_util.h"
#include "authselect.h"

static char *
append_line_to_buffer(char *buf,
                      const char *line,
                      size_t *bufsize)
{
    char *tmpbuf;
    size_t linelen;


    linelen = strlen(line);

    if (buf == NULL) {
        /* Terminating zero. */
        *bufsize = 1;
    }

    /* line + \n */
    tmpbuf = realloc_array(buf, char, *bufsize + linelen + 1);
    if (tmpbuf == NULL) {
        free(buf);
        return NULL;
    }
    buf = tmpbuf;

    memcpy(buf + *bufsize - 1, line, linelen);
    *bufsize += linelen;

    buf[*bufsize - 1] = '\n';
    *bufsize += 1;

    buf[*bufsize - 1] = '\0';

    return buf;
}

static errno_t
write_file(const char *path,
           const char *content)
{
    FILE *file = NULL;
    mode_t oldmask;
    size_t written;
    size_t len;
    errno_t ret;

    if (path == NULL || path[0] == '\0') {
        return EINVAL;
    }

    INFO("Creating file %s", path);

    /* Do not create an empty file if no content is given. */
    if (content == NULL) {
        content = "";
    }

    oldmask = umask(0644);

    file = fopen(path, "w");
    if (file == NULL) {
        ret = errno;
        ERROR("Unable to open file %s [%d]: %s", path, ret, strerror(ret));
        goto done;
    }

    len = strlen(content) + 1;
    written = fwrite(content, sizeof(char), len, file);
    if (written != len) {
        ret = errno;
        ERROR("Unable to write to a file [%d]: %s", ret, strerror(ret));
        goto done;
    }

    ret = chmod(path, 0644);
    if (ret != 0) {
        ret = errno;
        ERROR("Unable to chmod file [%d]: %s", ret, strerror(ret));
        goto done;
    }

    ret = EOK;

done:
    umask(oldmask);
    if (file != NULL) {
        fclose(file);
    }

    if (ret != EOK) {
        unlink(path);
    }

    return ret;
}

static errno_t
write_dconf(struct authselect_files *files)
{
    const char *line;
    char *buf;
    size_t bufsize;
    errno_t ret;

    buf = append_line_to_buffer(NULL, "# Generated by authselect", &bufsize);
    if (buf == NULL) {
        return ENOMEM;
    }

    buf = append_line_to_buffer(buf, "[org/gnome/login-screen]", &bufsize);
    if (buf == NULL) {
        return ENOMEM;
    }

    line = "enable-smartcard-authentication=false";
    if (files->smartcardauth != NULL && files->smartcardauth[0] != '\0') {
        line = "enable-smartcard-authentication=true";
    }
    buf = append_line_to_buffer(buf, line, &bufsize);
    if (buf == NULL) {
        return ENOMEM;
    }

    line = "enable-fingerprint-authentication=false";
    if (files->fingerprintauth != NULL && files->fingerprintauth[0] != '\0') {
        line = "enable-fingerprint-authentication=true";
    }
    buf = append_line_to_buffer(buf, line, &bufsize);
    if (buf == NULL) {
        return ENOMEM;
    }

    ret = write_file(AUTHSELECT_DCONF_FILE, buf);
    free(buf);

    return ret;
}

static errno_t
write_generated_files(struct authselect_profile *profile,
                      const char **optional)
{
    struct authselect_files *files;
    errno_t ret;

    ret = authselect_files_generate(profile, optional, &files);
    if (ret != EOK) {
        ERROR("Unable to generate content [%d]: %s", ret, strerror(ret));
        return ret;
    }

    ret = write_file(PATH_SYSTEM, files->systemauth);
    if (ret != EOK) {
        goto done;
    }

    ret = write_file(PATH_PASSWORD, files->passwordauth);
    if (ret != EOK) {
        goto done;
    }

    ret = write_file(PATH_FINGERPRINT, files->fingerprintauth);
    if (ret != EOK) {
        goto done;
    }

    ret = write_file(PATH_SMARTCARD, files->smartcardauth);
    if (ret != EOK) {
        goto done;
    }

    ret = write_file(PATH_NSSWITCH, files->nsswitch);
    if (ret != EOK) {
        goto done;
    }

    ret = write_dconf(files);
    if (ret != EOK) {
        goto done;
    }

done:
    authselect_files_free(files);

    return ret;
}

static errno_t
write_symbolic_links()
{
    struct authselect_symlink symlinks[] = SYMLINK_FILES;
    mode_t oldmask;
    errno_t ret;
    int i;

    oldmask = umask(0644);

    for (i = 0; symlinks[i].name != NULL; i++) {
        INFO("Creating symbolic link %s to %s",
             symlinks[i].name, symlinks[i].dest);

        ret = unlink(symlinks[i].name);
        if (ret != 0 && errno != ENOENT) {
            ret = errno;
            ERROR("Unable to overwrite file %s [%d]: %s",
                  symlinks[i].name, ret, strerror(ret));
            goto done;
        }

        ret = symlink(symlinks[i].dest, symlinks[i].name);
        if (ret != 0) {
            ret = errno;
            ERROR("Unable to create symbolic link %s [%d]: %s",
                  symlinks[i].name, ret, strerror(ret));
            goto done;
        }
    }

    ret = EOK;

done:
    umask(oldmask);

    return ret;
}

static errno_t
write_config(const char *profile_id,
             const char **optional)
{
    char *buf;
    size_t bufsize;
    errno_t ret;
    int i;

    buf = append_line_to_buffer(NULL, profile_id, &bufsize);
    if (buf == NULL) {
        return ENOMEM;
    }

    for (i = 0; optional[i] != NULL; i++) {
        buf = append_line_to_buffer(buf, optional[i], &bufsize);
        if (buf == NULL) {
            return ENOMEM;
        }
    }

    ret = write_file(PATH_CONFIG_FILE, buf);
    free(buf);

    return ret;
}

static errno_t
authselect_activate_profile(struct authselect_profile *profile,
                            const char **optional)
{
    errno_t ret;

    ret = write_generated_files(profile, optional);
    if (ret != EOK) {
        goto done;
    }

    ret = write_config(profile->id, optional);
    if (ret != EOK) {
        goto done;
    }

    ret = write_symbolic_links();
    if (ret != EOK) {
        goto done;
    }

done:
    return ret;
}

_PUBLIC_ int
authselect_activate(const char *profile_id,
                    const char **optional,
                    bool force_override)
{
    struct authselect_profile *profile;
    bool symlink_exist;
    bool is_valid;
    errno_t ret;

    INFO("Trying to activate profile [%s]", profile_id);

    profile = authselect_profile(profile_id);
    if (profile == NULL) {
        ERROR("Unable to find profile %s", profile_id);
        return ENOENT;
    }

    if (force_override) {
        INFO("Enforcing activation!");
        ret = authselect_activate_profile(profile, optional);
        goto done;
    }

    /* First, check that current configuration is valid. */
    ret = authselect_check_conf(&is_valid);
    if (ret != EOK && ret != ENOENT) {
        ERROR("Unable to check configuration [%d]: %s", ret, strerror(ret));
        goto done;
    }

    if (!is_valid) {
        ERROR("Unexpected changes to the configuration were detected and no "
              "overwrite option was specified.");
        ret = EINVAL;
        goto done;
    }

    /* If no configuration is present, check for existing files. */
    if (ret == ENOENT) {
        ret = authselect_check_symlinks_presence(&symlink_exist);
        if (ret != EOK) {
            ERROR("Unable to check for presence of symbolic links [%d]: %s",
                  ret, strerror(ret));
            goto done;
        }

        if (symlink_exist) {
            ERROR("File that needs to be overwritten was found and no "
                  "overwrite options was specified.");
            ret = EINVAL;
            goto done;
        }
    }

    ret = authselect_activate_profile(profile, optional);

done:
    if (ret != EOK) {
        ERROR("Unable to activate profile [%d]: %s", ret, strerror(ret));
    }

    authselect_profile_free(profile);

    return ret;
}
